#!/usr/bin/perl
########################################################
#
# This script is intended to provide a means for
# detecting changes made to files, via a regular
# comparison of SHA-1 hashes to an established
# "baseline". In this respect, it vaguely resembles
# the Tripwire product, hence the name. It now also
# captures all sorts of file information, though it's
# not doing a whole lot with it (except for the -paranoid
# feature).
#
# This script requires perl (duh), Berkeley DBM support
# and the Digest::SHA1 module, which you can get from
# CPAN in the DIGEST bundle. 
#
########################################################
########################################################
#
# Usage:
#
# ./slipwire [options] [hashfilename] [file_list]|file1 file2 file3...
#
# The options are as follows:
#	
#	-create		Initialize database file
#
#       -only           Index only the filenames given,
#                       ie, don't use a list of files.
#
# 	-compare	Compare SHA-1 hashes to 
#			baseline state.
#
#       -paranoid       Compare inode, last modify time and
#                       filesize in addition to the SHA-1
#                       hashes.
#
#
# hashfilename is the name of the Berkeley DBM file created
# for storing the filenames and SHA-1 hashes.
#
# file_list is a text file containing a list of
# paths that you want to index.
#
#
# example:
#
# slipwire.pl -create test.dbm list_example.txt
#
#  ...creates test.dbm and recursively generates 
#  entries and SHA-1 hashes for all directories listed
#  in list_example.txt, which is an ascii file containing
#  one path per line. See the example file for details.
#
# slipwire.pl -only test.dbm file1 file2 file3 file4
# ...creates foo.dbm but only indexes file1, file2,
# file3 and file4. 
#
# slipwire.pl -compare test.dbm list_example.txt
# ...checks the SHA hashes of the files in 
# list_example.txt against the current entries
# in test.dbm, alerting you to changes.
#
# slipwire.pl -compare test.dbm file1 file2 file3 file4
# ...checks the SHA hashes of file1, file2,file3, file4
#
# There isn't much in the way of error-checking in this
# yet, but I thought that this might be useful to someone
# even in its current form. I'm still working on it.
#
# For perl assistance and general feedback, 
# my sincere thanks go to:
#
# Mark Wolgemuth (mark@node.to)
# Francisco Roque (franisco@perauno.org)
#
# This script can be redistributed and/or modified under 
# the same term as perl itself.
#
# Cheers,
#
# James Quinby 
# jquinby@node.to
# http://packet.node.to
#
#######################################################
#######################################################

use DB_File;
use FileHandle;
use Digest::SHA1 qw(sha1_base64);
require "find.pl";

if  ($> != 0){
  print "\nSorry, you must be root to run this program.\n\n";
  exit;
}


if ($ARGV[0] =~ /-create/) {
  create();
  exit;
}

if ($ARGV[0] =~ /-compare/) {
  compare();
  exit;
}

if ($ARGV[0] =~ /-only/) {
  only();
  exit;
}


if ($ARGV[0] =~ /-paranoid/) {
  paranoid();
  exit;
}



if ($ARGV[0] =~ //) {
  usage();
  exit;
}


sub create {
 
  my $file_list = new FileHandle;
  my $i = 0;

  $file_list->open($ARGV[2]); 
  $db = tie  %hashfile, "DB_File", $ARGV[1];

  my $file_list_line;

  while ($file_list_line = $file_list->getline()){
    chop $file_list_line;

    print "\nCreating hashes for $file_list_line...\n\n";
    

    find($file_list_line);

    sub wanted {

      $i++;

      open (FILEH, $name);

      my $key = $name;
  
      $ctx = Digest::SHA1->new;
      $ctx->addfile(FILEH);
      $sha = $ctx->b64digest; 

      my @fileinfo;

      @fileinfo = lstat($name);

      # Each file gets a pipe-delimited entry, with format as follows:
      #
      # field 0 = SHA-1 hash
      # field 1 = device number of filesystem
      # field 2 = inode number
      # field 3 = file mode (type/permissions)
      # field 4 = number of hard links to file
      # field 5 = uid of the file
      # field 6 = gid of the file
      # field 7 = filesize
      # field 8 = last modify time (since epoch)
      #
      # See the perl "stat" command or man stat() for more information.
      # -JQ

      $entry = "$sha|$fileinfo[0]|$fileinfo[1]|$fileinfo[2]|$fileinfo[3]|$fileinfo[4]|$fileinfo[5]|$fileinfo[7]|$fileinfo[9]";

      print "$key|$entry\n";	 

      $db->put($key, $entry);
      close FILEH;
    }
  }

print "\nHash database created successfully. $i files entered.\n";
  
}


sub compare {  
  print "Comparing files against SHA signatures...\n\n";

  my $file_list = new FileHandle;
  my $i;
  my $j = 0;

  $db = tie  %hashfile, "DB_File", $ARGV[1];

  $file_list->open($ARGV[1]);

  my $file_list_line;

  while (($key, $entry) = each %hashfile) {     

    open (FILEH, $key);

    @fileinfo = split /\|/, $entry;

    $ctx = Digest::SHA1->new;
    $ctx->addfile(FILEH);
    $new_value = $ctx->b64digest;
 	
    $i++;  

    # Check for SHA changes
 
    if ($new_value ne $fileinfo[0]) {

      print "ALERT! FILE CHANGED: $key\n";
      print "old SHA signature: " . $fileinfo[0] . "\n";
      print "new SHA signature: " . $new_value . "\n\n";
      
      $j++;
	
    }
  }
  print "\nComparison complete. $i files checked with $j alerts.\n\n";
}


sub only {

  my $file_list;
  my $i = 0;
  my @file_list = splice(@ARGV, 2);

  $db = tie  %hashfile, "DB_File", $ARGV[1];
  print "\nCreating hashes...\n\n";

  while ($file_list = shift @file_list){

    $i++;
    
    my $key = $file_list;
    chomp $key;  
    
    my $value = `md5 -q $key`;
    chomp $value;

    print "$key: $value\n";	 	
    $db->put($key, $value);
  }
  print "\nHash database created successfully. $i files entered.\n";
}


sub paranoid {

  print "Comparing files against SHA signatures, inode, last-modified and filesize changes...\n\n";

  my $file_list = new FileHandle;
  my $i = 0;
  my $j = 0;
  my @oldfileinfo;
  my @newfileinfo;
  my $new_value;
  my $db;

  $db = tie  %hashfile, "DB_File", $ARGV[1];

  $file_list->open($ARGV[1]);

  while (($key, $entry) = each %hashfile) {     

    open (FILEH, $key);

    @oldfileinfo = split /\|/, $entry;
    @newfileinfo = lstat($key);

    $ctx = Digest::SHA1->new;
    $ctx->addfile(FILEH);
    $new_value = $ctx->b64digest;
 	
    # Check for SHA changes

    $i++;  
 
    if ($new_value ne $oldfileinfo[0]) {
      print "SHA CHANGE: $key\n";
      print "\told SHA signature: " . $oldfileinfo[0] . "\n";
      print "\tnew SHA signature: " . $new_value . "\n\n";
      
      $j++;
    }

    # Check for inode changes

    if ($oldfileinfo[2] ne $newfileinfo[1]){
      print "INODE CHANGE: $key\n";
      print "\told INODE: " . $oldfileinfo[2] . "\n";
      print "\tnew INODE: " . $newfileinfo[1] . "\n\n";

      $j++;


      if ($newfileinfo[1] eq ""){
	print "=> Check for deletion or replacement with symlink!\n\n";
      }
    }

    # Check for filesize changes
	
    if ($oldfileinfo[7] ne $newfileinfo[7]){
      print "SIZE CHANGE: $key\n";
      print "\told SIZE: " . $oldfileinfo[7] . "\n";
      print "\tnew SIZE: " . $newfileinfo[7] . "\n\n";
      
      $j++;


	if ($newfileinfo[7] eq ""){
	  print "=> Check for deletion!\n\n";
	}
    }

    # Check for, well, last-modified changes

    if ($oldfileinfo[8] ne $newfileinfo[9]){
      print "LAST-MODIFY TIME CHANGE: $key\n";
      print "\told TIME: " . $oldfileinfo[8] . "\n";
      print "\tnew TIME: " . $newfileinfo[9] . "\n\n";
      
      $j++;


	if ($newfileinfo[9] eq ""){
	  print "=> Check for deletion!\n\n";
	}
    }
  }
  print "\nComparison complete. $i files checked with $j alerts.\n\n";

}

sub usage {
  print "\n";
  print "Usage: $0 [options] hashfile [file-list|file1 file2 file3...]\n\n";
  print "Options:\n\n";
  print "-create\nInitialize the hash.dbm database\n\n";
  print "-only\nCreate index with supplied filenames only\n\n";
  print "-compare\nCheck the filesystems listed in [file-list] for SHA changes.\n\n";
  print "-paranoid\nCheck the filesystems listed in [file-list] for SHA, inode, size or date changes.\n\n";
  print "\n";
  print "hashfile is a DBM file created and used for storing the file and SHA information.";
  print "\n";
  print "file-list is a text file containing paths and/or filenames to be indexed.\n";
  print "\n";
}
